library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.math_pack.all;

entity gpdrc_top is
    generic(
        PRM 		    : integer := 4;                                   -- pocet pripojenych PRM
        VALUE		    : integer := 184320; -- 1843200                   -- maximalni hodnota citace v Safety Window 
        BITS		    : integer := 24;                                  -- pocet bitu citace v Safety Window
        SIZE_RAM 	    : integer := 20 	                              -- sirka adresove sbernice do flash
    );	
    port(
        CLK 		    : in std_logic;                                 -- hodinovy signal
        RST		    : in std_logic;                                 -- globalni reset
        PRM_ERR 	    : in std_logic_vector(PRM-1 downto 0);          -- sbernice pripojena k jednotlivym PRM
        OE                  : out std_logic;                                -- kontrolni signal flash Output Enable
        CE                  : out std_logic;                                -- kontrolni signal flash Chip Enable
        WE                  : out std_logic;                                -- kontrolni signal flash Write Enable
        START_RECONFIG      : out std_logic;                                -- impulz pri startu rekonfigurace
        END_RECONFIG        : out std_logic;                                -- impulz pri konci rekonfigurace
        HARD_ERR	    : out std_logic;                                -- signal oznacujici, ze v PRM je trvala porucha 
        SYNC		    : out std_logic;                                -- synchronizacni signal po Safety Window
        FLASH_ADDR          : out std_logic_vector(SIZE_RAM-1 downto 0);    -- adresa bytu ke cteni do flash pameti
        FLASH_DATA_IN       : in std_logic_vector(7 downto 0);              -- 8bit data z flash
        DATA_TO_ICAP        : out std_logic_vector(31 downto 0);            -- 32bit data z FIFO na ICAP rozhrani
        WRITE_TO_ICAP       : out std_logic;                                -- povoleni zapisu na ICAP aktivni v log. 0
        ICAP_CLK            : out std_logic;                                -- hodinovy signal pro ICAP (prozatim stejny jak CLK)
        PRM_INDEX           : out std_logic_vector(log2(PRM)-1 downto 0)    -- index aktualne rekonfigurovaneho PRM
    ); 
end gpdrc_top;



architecture struct of gpdrc_top is

--==============Round Robin============================================
component roundrobin is                                                -- vybira pouze jeden PRM s poruchou a vraci jeho index v kodovani 1 z N
    generic(
        PRM: integer                                                   -- pocet pripojenych PRM
    );
    port(
        clk	: in std_logic;                                        -- hodinovy signal
        rst	: in std_logic;                                        -- reset
        enable	: in std_logic;                                        -- povoleni posunu 
        enable_data_out : in std_logic;                                -- povoleni vystaveni vybraneho prvku na vystup
        data_in	        : in std_logic_vector(PRM-1 downto 0);         -- vstupni sbernice (pripojeni sbernice od PRM)
        data_out	: out std_logic_vector(PRM-1 downto 0);        -- vystupni sbernice (aktivni maximalne jeden vybrany signal)
        done            : out std_logic                                -- signal oznacujici konec cyklu (kontrolovan nejnizsi bit)
    );
end component;
----------------------------------------------------------------------

--==============Error Decoder=========================================
component errdec is                                                    -- konvertuje vstupni data z kodovani 1 z N na index
    generic(
        PRM: integer                                                   -- pocet pripojenych PRM
	  );
    port( 
        errin  : in std_logic_vector(PRM-1 downto 0);                    -- vstupni sbernice (pripojeni sbernice z vystupu Round Robin)
        stop   : out std_logic;                                        -- signal pro automat, pokud byla vybran PRM s poruchou
        errout : out std_logic_vector(log2(PRM)-1 downto 0)          -- vystupni sbernice s indexem PRM prevedenym z kodu 1 z N na vstupu 
    ); 
end component;
----------------------------------------------------------------------	
	
--==============Hard Error Unit=========================================
component harderr is                                                   -- kontroluje zda se PRM s poruchou nerekonfiguruje opakovane
    generic(
        PRM: integer                                                   -- pocet pripojenych PRM	 
	  );
    port( 
        clk         : in std_logic;                                    -- hodinovy signal
        rst         : in std_logic;                                    -- reset
        rrerr_in    : in std_logic_vector(PRM-1 downto 0);             -- index aktualne rekonfigurovaneho PRM s poruchou v kdu 1 z N (pripojeno na vystup Round Robin)
        harderr     : out std_logic;                                   -- oznacuje, zda se PRM s poruchou neopakuje v druhem cyklu Round Robin po sobe (log. 1 = trvala porucha)
        refresh     : in std_logic;                                    -- prikaz k aktualizaci registru minule rekonfigurovanych PRM hodnotou registru s aktualne rekonfigurovanymy PRM (na konci cyklu Round Robin)
        clear       : in std_logic;                                    -- vynulovani registru aktualne rekonfigurovanych PRM
        enable      : in std_logic                                     -- povoleni vystaveni signalu harderr (pro automat)
    ); 
end component;
----------------------------------------------------------------------

--==============Safety Window Unit=====================================
component safetyw is                                                   -- jednotka pro synchronizaci PRM tvorena citacem, ten po preteceni aktivuje signal sync
    generic(	
        VALUE: integer;                                                -- maximalni hodnota citace v Safety Window 
	      BITS: integer                                                  -- pocet bitu citace v Safety Window
	  );
    port( 
        clk      : in std_logic;                                       -- hodinovy signal
        rst      : in std_logic;	                                     -- reset
        enable   : in std_logic;                                       -- povoleni citani     
        sync     : out std_logic                                       -- signal informujici o dokonceni citani do dane hodnoty
    ); 
end component;
----------------------------------------------------------------------

--==============Address Counter=====================================
component adrcnt is                                                    -- adresivy citac citajici od pocatku do konce bitstreamu a vracejici adresu aktualniho bytu z neho 
    generic(
        PRM: integer;                                                  -- pocet pripojenych PRM	
	      SIZE_RAM : integer 	                                           -- sirka adresove sbernice do flash
	  );
	  port( 
        clk       : in std_logic;                                      -- hodinovy signal
        rst       : in std_logic;	                                     -- reset

        start     : in std_logic;                                      -- signal nastaveni pro nastaveni pocatecni adresy bitstreamu a povoleni citani 
        enable    : in std_logic;                                      -- povoleni citani
        stop      : out std_logic;                                     -- signal informujici automat o docitani na posledni adresu v danem bitstreamu
      
        adr_start : in std_logic_vector(SIZE_RAM-1 downto 0);            -- adresa pocatku daneho bitstreamu (z LUT)
        adr_end   : in std_logic_vector(SIZE_RAM-1 downto 0);            -- adresa konce daneho bitstreamu (z LUT)

        adr_bit   : out std_logic_vector(SIZE_RAM-1 downto 0);           -- adresa aktualniho bytu ke cteni pro Flash Controller
        adr_valid : out std_logic                                      -- signal oznacujici nastaveni adresy
    ); 
end component;
----------------------------------------------------------------------

--==============LOOK UP=====================================
component lookup is                                                    -- vyhledavaci tabulka vracejici podle indexu pocatecni a koncovou adresu bitstreamu
    port (
	      a: IN std_logic_VECTOR(7 downto 0);                            -- index PRM
	      spo: OUT std_logic_VECTOR(39 downto 0)                         -- pocatecni (39 az 20) a koncova (19 az 0) adresa bitstreamu odpovidajiciho PRM na vstupu
    );
end component;
----------------------------------------------------------------------

--==============FIFO=====================================
component fifo is                                                      -- vyrovnavaci pamet mezi flash a ICAP
	  port (
	      rst: in std_logic;                                             -- reset
	      wr_clk: in std_logic;                                          -- hodiny pro zapis
	      rd_clk: in std_logic;                                          -- hodiny pro cteni (prozatim stejne jako pro zapis)
	      din: in std_logic_vector(7 downto 0);                          -- 8bit data z Flash do FIFO 
	      wr_en: in std_logic;                                           -- povoleni zapisu
	      rd_en: in std_logic;                                           -- povoleni cteni
	      dout: out std_logic_vector(31 downto 0);                       -- 32bit data z FIFO do ICAP 
	      full: out std_logic;                                           -- FIFO plne (nepouzit)
	      empty: out std_logic                                           -- FIFO prazdne
    );
end component;

--==============ECC parity=====================================
component frame_parity is                                              -- kontrola parity daneho framu bitstreamu (prozatim odpojene)
    port(
        clk : in std_logic;                                            -- hodinovy signal
        rst : in std_logic;                                            -- reset
        start : in std_logic;                                          -- signal oznacujici prvni byte framu a tedy pocatku kontroly (inicializace XORu)
        data_in : in std_logic_vector(7 downto 0);                     -- vstupni data z flash
        data_valid : in std_logic;                                     -- data platne 
        parity_ok : out std_logic                                      -- signal oznacujici aktualni stav parity (log.1 = parita v poradku)
    );
end component;
----------------------------------------------------------------------

--==============FSM=====================================
component fsm is                                                       -- automat ridici GPDRC
    port( 
        clk        : in std_logic;                                     -- hodinovy signal
        rst        : in std_logic;                                     -- reset
        err        : in std_logic;                                     -- nalezena chyba, rekonfigurace (z Errdec)
      
        start_reconfig : out std_logic;                                -- impuls pri startu rekonfigurace PRM
        end_reconfig : out std_logic;                                  -- impuls pri konci rekonfigurace PRM
        enable_safety_w : out std_logic;                               -- povoleni citani v Safety Window
        sync       : in std_logic;                                     -- dokonceni citani ze Safety Window
       
        clear_input_reg : out std_logic;                               -- vynulovani zachytneho registru na vstupu PRM_ERROR
                
        refresh_hard : out std_logic;                                  -- aktualizace registru v jednotce HardErr
        clear_hard  : out std_logic;                                   -- vynulovani registru v jednotce HardErr
        enable_hard : out std_logic;                                   -- povoleni vystaveni signal o trvale poruse PRM (HARD_ERR)
            
        parity_start : out std_logic;                                  -- informace o zacatku framu do ECC
        parity_ok  : in std_logic;                                     -- informace o stavu parity z ECC
      
        start_cnt  : out std_logic;                                    -- inicializace a povoleni citani citace AdrCnt
        enable_cnt : out std_logic;                                    -- povoleni citani citace AdrCnt
        stop_cnt   : in std_logic;                                     -- informace z AdrCnt o dokonceni citani na posledni byte bitstreamu
      
        read_fifo  : out std_logic;                                    -- povoleni vycitani dat z FIFO
        empty_fifo : in std_logic;                                     -- informace o prazdnosti FIFA
                          
        enable_rr  : out std_logic;                                    -- povoleni cinnosti Round Robin
        done_rr    : in std_logic;                                     -- informace o dokonceni jednoho cyklu z Round Robin

        data_valid : in std_logic;                                     -- data z flash platna
        data       : in std_logic_vector(7 downto 0)                   -- data z flash
    ); 
end component;
----------------------------------------------------------------------

--==============Flash============================================
component flash_controller is                                          -- kontroler flash pameti (AMIC A29010)
    generic(
	      SIZE_RAM : integer 	                                           -- sirka adresove sbernice do flash 
	  );
    port (
	      oe_n : out std_logic;                                          -- kontrolni signal flash Output Enable
        we_n : out std_logic;                                          -- kontrolni signal flash Write Enable
	      ce_n : out std_logic;                                          -- kontrolni signal flash Chip Enable
	      addr : out std_logic_vector(SIZE_RAM-1 downto 0);                -- adresa bytu ke cteni z kontroleru do flash
	      io : in std_logic_vector(7 downto 0);                          -- data z flash do kontroleru
	      rd : in std_logic;                                             -- signal o potrebe cteni z automatu do kontroleru
	      address : in std_logic_vector(SIZE_RAM-1 downto 0);	             -- adresa bytu ke cteni do kontroleru
	      reset : in std_logic;                                          -- reset
	      data : out std_logic_vector(7 downto 0);                       -- precteny byte z kontroleru do dalsich jednotek
	      data_valid : out std_logic;                                    -- informace o platnosti precteneho bytu z kontroleru
	      clk : in std_logic	                                           -- hodinovy signal (maximalne 10 MHz, prozatim spojeno s clk celeho GPDRC)                 
	  );
end component;
----------------------------------------------------------------------

--==============Input Register======================================
component input_register is                                            -- zachytny regist pripojeny na sbernici PRM_ERROR na vstupu GPDRC
    generic(
        PRM: integer	                                                 -- pocet pripojenych PRM	
	  );
    port( 
        clk         : in std_logic;                                    -- hodinovy signal
        rst         : in std_logic;                                    -- reset
        clear	      : in std_logic;                                    -- nulovani registru
        data_in	    : in std_logic_vector(PRM-1 downto 0);               -- vstupni sbernice (pripojeno k PRM_ERROR)
        data_out	  : out std_logic_vector(PRM-1 downto 0)               -- vystupni sbernice (pripojeno k Round Robin)
    ); 
end component;

----------------------------------------------------------------------



-------------------------------------------------------------------

signal decoder_errin : std_logic_vector(PRM-1 downto 0);
signal adr_data : std_logic_vector(log2(PRM)-1 downto 0);
signal hard_done, sw_err, refresh_hard_sig,  hard_err_sig : std_logic;
signal enable_hard_sig, clear_hard_sig : std_logic;
signal enable_cnt_sig, stop_cnt_sig : std_logic;
signal parity_start_sig, parity_ok_sig : std_logic;
signal clear_input_reg_sig : std_logic;
signal adr_start,adr_end : std_logic_vector(SIZE_RAM-1 downto 0);
signal adr_valid_sig, data_valid_sig : std_logic;
signal adr_cnt_flash : std_logic_vector(SIZE_RAM-1 downto 0);
signal data_to_icap_sig : std_logic_vector(31 downto 0);

signal enable_rr_sig, start_cnt_sig : std_logic;

signal empty_sig : std_logic;
signal adr_data_wide : std_logic_vector(7 downto 0);
signal data_to_fifo : std_logic_vector(7 downto 0);
signal data_from_fifo_enable : std_logic;
signal spo_sig : std_logic_vector(39 downto 0);

signal sync_sig, enable_safety_w_sig : std_logic;
signal prm_errors : std_logic_vector(PRM-1 downto 0);
signal icap_enable_sig : std_logic;
 

-----------------------------------------------------------
begin

    Input_Reg : input_register
        generic map(
            PRM=>PRM	 
	      )
        port map( 
            clk => CLK,
            rst => RST,
            clear => clear_input_reg_sig,             
            data_in => PRM_ERR,
            data_out => prm_errors
        );
         
    Robin : roundrobin
        generic map(
            PRM=>PRM
		    )
        port map(
            clk	=> CLK,
            rst	=> RST,
            --stop => sw_err,
            enable => enable_rr_sig,
            enable_data_out => enable_hard_sig,
            data_in	=> prm_errors,
            data_out	=> decoder_errin,
            done      => hard_done
        );

    ErrDec_FU : errdec
        generic map(
            PRM=>PRM
		    )
        port map(
            errin	=> decoder_errin,
            stop	=> sw_err,
            errout	=> adr_data
        );

    HardError : harderr
        generic map(
            PRM=>PRM
		    )
        port map(
            clk	=> CLK,
            rst	=> RST,
            rrerr_in	=> decoder_errin,
            refresh	=> refresh_hard_sig,
            enable => enable_hard_sig,
            clear => clear_hard_sig,
            harderr	=> hard_err_sig
        );

    SafetyW_FU : safetyw
      	generic map(
		        VALUE=>VALUE,
		        BITS=>BITS
		    )
        port map(
            clk	=> CLK,
            rst	=> RST,
            enable	=> enable_safety_w_sig,
            sync	=> sync_sig
        );


    adr_data_wide(log2(PRM)-1 downto 0) <= adr_data;
    adr_data_wide(7 downto log2(PRM)) <= (others => '0');    
    
    adr_end <= spo_sig(19 downto 0);
    adr_start <= spo_sig(39 downto 20);

    adr_lookup_fu: lookup
        port map(
            a =>  adr_data_wide,
            spo => spo_sig
        );


    AddrCnt_FU : adrcnt
      	generic map(
            PRM=>PRM,
		        SIZE_RAM=>SIZE_RAM
		    )
        port map(
            clk	=> CLK,
            rst	=> RST,

            start => start_cnt_sig,
            enable => enable_cnt_sig,
            stop => stop_cnt_sig,
      
            adr_start	=> adr_start,
            adr_end	=> adr_end,

            adr_bit	=> adr_cnt_flash,
            adr_valid => adr_valid_sig
        );

    FIFO_FU : fifo
        port map( 
     	      rst => RST,
	          wr_clk => CLK, 
	          rd_clk => CLK, -- prozatim
	          din => data_to_fifo,
	          wr_en => data_valid_sig,
	          rd_en => data_from_fifo_enable,
	          dout => data_to_icap_sig,
	          full => open, 
	          empty => empty_sig
	      );

    ECC: frame_parity
        port map(
            clk => CLK,
            rst => RST,
            start => parity_start_sig,
            data_in => data_to_fifo,
            data_valid => data_valid_sig,
            parity_ok => parity_ok_sig
        );

    FSMa: fsm
        port map(
            clk	=> CLK,
            rst	=> RST,
            err => sw_err,  

            start_reconfig => START_RECONFIG,            
            end_reconfig => END_RECONFIG,
            start_cnt => start_cnt_sig,
            enable_cnt => enable_cnt_sig,
            stop_cnt => stop_cnt_sig,
      
            read_fifo => data_from_fifo_enable,
            empty_fifo => empty_sig,
            done_rr => hard_done,                      
            enable_rr => enable_rr_sig,
        
            clear_input_reg => clear_input_reg_sig,
            refresh_hard => refresh_hard_sig,
            enable_hard => enable_hard_sig,
            clear_hard => clear_hard_sig,
            
            enable_safety_w => enable_safety_w_sig,
            sync => sync_sig,
        
            parity_start => parity_start_sig,
            parity_ok => parity_ok_sig,
      
            data_valid => data_valid_sig,
            data => data_to_fifo
        );

    FLASH_CTRL: flash_controller
        generic map(
	          SIZE_RAM => SIZE_RAM
	      )
        port map(
	          oe_n => OE,
            we_n => WE,
	          ce_n => CE,
	          addr => FLASH_ADDR,
	          io => FLASH_DATA_IN,
	
	          rd => adr_valid_sig,
	          address => adr_cnt_flash,
	          reset => RST,
	          data => data_to_fifo,
	          data_valid => data_valid_sig,
	          clk => CLK

	      );
	  
    icap_enable : process(clk, rst, data_from_fifo_enable, empty_sig)
    begin
        if(rst = '1') then  
            icap_enable_sig <= '0';
        else  
            if(rising_edge(clk)) then
                if(empty_sig = '1') then
                    icap_enable_sig <= '0';
                elsif(data_from_fifo_enable = '1') then
                    icap_enable_sig <= '1';
                end if;
            end if;
        end if;
    end process;
	  

    WRITE_TO_ICAP <= not icap_enable_sig;
    DATA_TO_ICAP <= data_to_icap_sig;
	  ICAP_CLK <= clk;
	  
	  SYNC <= sync_sig;
    PRM_INDEX <= adr_data;
    HARD_ERR <= hard_err_sig;
    
end struct;
